//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "CFFT.h"

#pragma link "config.obj"
#pragma link "executor.obj"
#pragma link "fftwf77.obj"
#pragma link "fftwnd.obj"
#pragma link "fni_1.obj"
#pragma link "fni_10.obj"
#pragma link "fni_11.obj"
#pragma link "fni_12.obj"
#pragma link "fni_13.obj"
#pragma link "fni_14.obj"
#pragma link "fni_15.obj"
#pragma link "fni_16.obj"
#pragma link "fni_2.obj"
#pragma link "fni_3.obj"
#pragma link "fni_32.obj"
#pragma link "fni_4.obj"
#pragma link "fni_5.obj"
#pragma link "fni_6.obj"
#pragma link "fni_64.obj"
#pragma link "fni_7.obj"
#pragma link "fni_8.obj"
#pragma link "fni_9.obj"
#pragma link "fn_1.obj"
#pragma link "fn_10.obj"
#pragma link "fn_11.obj"
#pragma link "fn_12.obj"
#pragma link "fn_13.obj"
#pragma link "fn_14.obj"
#pragma link "fn_15.obj"
#pragma link "fn_16.obj"
#pragma link "fn_2.obj"
#pragma link "fn_3.obj"
#pragma link "fn_32.obj"
#pragma link "fn_4.obj"
#pragma link "fn_5.obj"
#pragma link "fn_6.obj"
#pragma link "fn_64.obj"
#pragma link "fn_7.obj"
#pragma link "fn_8.obj"
#pragma link "fn_9.obj"
#pragma link "ftwi_10.obj"
#pragma link "ftwi_16.obj"
#pragma link "ftwi_2.obj"
#pragma link "ftwi_3.obj"
#pragma link "ftwi_32.obj"
#pragma link "ftwi_4.obj"
#pragma link "ftwi_5.obj"
#pragma link "ftwi_6.obj"
#pragma link "ftwi_64.obj"
#pragma link "ftwi_7.obj"
#pragma link "ftwi_8.obj"
#pragma link "ftwi_9.obj"
#pragma link "ftw_10.obj"
#pragma link "ftw_16.obj"
#pragma link "ftw_2.obj"
#pragma link "ftw_3.obj"
#pragma link "ftw_32.obj"
#pragma link "ftw_4.obj"
#pragma link "ftw_5.obj"
#pragma link "ftw_6.obj"
#pragma link "ftw_64.obj"
#pragma link "ftw_7.obj"
#pragma link "ftw_8.obj"
#pragma link "ftw_9.obj"
#pragma link "generic.obj"
#pragma link "malloc.obj"
#pragma link "planner.obj"
#pragma link "putils.obj"
#pragma link "rader.obj"
#pragma link "timer.obj"
#pragma link "twiddle.obj"
#pragma link "wisdom.obj"
#pragma link "wisdomio.obj"

#pragma package(smart_init)

#include <math.h>

//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//

static inline void ValidCtrCheck(TCFFT *)
{
        new TCFFT(NULL);
}
//---------------------------------------------------------------------------
__fastcall TCFFT::TCFFT(TComponent* Owner)
        : TComponent(Owner)
{
  FFTSize = 0;
  CurrentWeightingSize = 0;
  CurrentFFTWeighting  = NoWeighting;
  CurrentTaylorN = 0;
  CurrentSidelobeLevel = 0;
  Weighting = NULL;
}

//---------------------------------------------------------------------------
__fastcall TCFFT::~TCFFT(){
  if(Weighting) free(Weighting);
}

//---------------------------------------------------------------------------
namespace Cfft
{
        void __fastcall PACKAGE Register()
        {
                 TComponentClass classes[1] = {__classid(TCFFT)};
                 RegisterComponents("Samples", classes, 0);
        }
}
//---------------------------------------------------------------------------

/*
     fftw_plan p;
     fftw_complex in[N], out[N];
     p = fftw_create_plan(N, FFTW_FORWARD, FFTW_ESTIMATE);
     fftw_one(p, in, out);
     fftw_destroy_plan(p);
*/

bool __fastcall TCFFT::CreatePlan(int N){
  if(FGPL){
    fftw_direction fftdir;
    if(FFFTDirection == Forward){
      fftdir = FFTW_FORWARD;
    } else {
      fftdir = FFTW_BACKWARD;
    }
    int plantype;
    if(FPlanType == Estimate){
      plantype = FFTW_ESTIMATE;
    } else {
      plantype = FFTW_MEASURE;
    }
    pfftplan = fftw_create_plan(N, fftdir, plantype);
  }
  FFTSize = N;
  return true;
}

//---------------------------------------------------------------------------
bool __fastcall TCFFT::DestroyPlan(){
  if(FGPL){
    fftw_destroy_plan(pfftplan);
  }
  return true;
}


//---------------------------------------------------------------------------
bool __fastcall TCFFT::fftw_fwd_bckwd(fftw_complex* in, fftw_complex* out){
  if(!CurrentWeightingValid()){
    if(!CreateWeighting()) return false;
  }
  if(FFTDirection == Backward){
    for(int i=0; i<FFTSize; i++){
      in[i].im = -in[i].im;
    }
  }
  if(FGPL){
    if(FFTWeighting != NoWeighting){
      fftw_complex *cin = (fftw_complex*)calloc(FFTSize, sizeof(fftw_complex));
      if(!cin){
        return false;
      }
      for(int i=0; i<FFTSize; i++){
        cin[i] = in[i];
      }
      for(int i=0; i<FDataSize; i++){
        cin[i].re *= Weighting[i];
        cin[i].im *= Weighting[i];
      }
      fftw_one(pfftplan, cin, out);
      free(cin);
    } else {
      fftw_one(pfftplan, in, out);
    }
  } else {
    Complex *cin = (Complex*)calloc(FFTSize, sizeof(Complex));
    if(!cin){
      return false;
    }
    for(int i=0; i<FFTSize; i++){
      cin[i] = Complex(in[i].re, in[i].im);
    }
    if(FFTWeighting != NoWeighting){
      for(int i=0; i<FDataSize; i++){
        cin[i] *= Weighting[i];
      }
    }
    ipFFT_slow(cin);
    for(int i=0; i<FFTSize; i++){
      out[i].re = cin[i].real();
      out[i].im = cin[i].imag();
    }
    free(cin);
  }
  if(FFTDirection == Backward){
    for(int i=0; i<FFTSize; i++){
      in[i].im = -in[i].im;
      out[i].im = -out[i].im;
    }
  }
  return true;
}


//---------------------------------------------------------------------------
bool __fastcall TCFFT::FFT(fftw_complex *in, fftw_complex *out){
  fftw_fwd_bckwd(in, out);
  Normalize(out);
  return true;
}

//---------------------------------------------------------------------------
bool __fastcall TCFFT::FFT(complex<double> *in, complex<double> *out){
  // For now at least, complex<double> is the same as fftw_complex
  fftw_fwd_bckwd((fftw_complex*)in, (fftw_complex*)out);
  Normalize((fftw_complex*)out);
  return true;
}

//---------------------------------------------------------------------------

bool __fastcall TCFFT::FFT(complex<float> *in, complex<float> *out){
  fftw_complex *fftin = (fftw_complex*)calloc(FFTSize, sizeof(fftw_complex));
  if(!fftin){
    return false;
  }
  fftw_complex *fftout = (fftw_complex*)calloc(FFTSize, sizeof(fftw_complex));
  if(!fftout){
    free(fftin);
    return false;
  }
  for(int i=0; i<FFTSize; i++){
    fftin[i].re = in[i].real();
    fftin[i].im = in[i].imag();
  }
  fftw_fwd_bckwd(fftin, fftout);
  Normalize((fftw_complex*)out);
  for(int i=0; i<FFTSize; i++){
    out[i] = complex<float>(fftout[i].re, fftout[i].im);
  }

  free(fftin);
  free(fftout);
  return true;
}


//---------------------------------------------------------------------------

bool __fastcall TCFFT::FFT(fftw_complex *in){
  fftw_complex *out = (fftw_complex*)calloc(FFTSize, sizeof(fftw_complex));
  if(!out){
    return false;
  }
  fftw_fwd_bckwd(in, out);
  Normalize(out);
  memcpy(in, out, FFTSize*sizeof(fftw_complex));
  free(out);
  return true;
}

//---------------------------------------------------------------------------

bool __fastcall TCFFT::FFT(complex<double> *in){
  // For now at least, complex<double> is the same as fftw_complex
  fftw_complex *out = (fftw_complex*)calloc(FFTSize, sizeof(fftw_complex));
  if(!out){
    return false;
  }
  fftw_fwd_bckwd((fftw_complex*)in, out);
  Normalize(out);
  memcpy(in, out, FFTSize*sizeof(complex<double>));
  free(out);
  return true;
}

//---------------------------------------------------------------------------

bool __fastcall TCFFT::FFT(complex<float> *in){
  fftw_complex *fftin = (fftw_complex*)calloc(FFTSize, sizeof(fftw_complex));
  if(!fftin){
    return false;
  }
  fftw_complex *fftout = (fftw_complex*)calloc(FFTSize, sizeof(fftw_complex));
  if(!fftout){
    free(fftin);
    return false;
  }
  for(int i=0; i<FFTSize; i++){
    fftin[i].re = in[i].real();
    fftin[i].im = in[i].imag();
  }
  fftw_fwd_bckwd(fftin, fftout);
  Normalize((fftw_complex*)fftout);
  for(int i=0; i<FFTSize; i++){
    in[i] = complex<float>(fftout[i].re, fftout[i].im);
  }
  free(fftin);
  free(fftout);
  return true;
}




//---------------------------------------------------------------------------

bool __fastcall TCFFT::Normalize(fftw_complex *out){
  if(FNormalization == None) return true;
  if(FFTSize < 1) return false;
  double scale = 1.0;
  float fsize = (float)FFTSize;
  float dsize = (float)FDataSize;
  bool brc = true;
  switch(FNormalization){
    case SqrtN:
      scale = sqrt(fsize);
    break;
    case N:
      scale = fsize;
    break;
    case NInv:
      scale = 1.0/fsize;
    break;
    case SqrtNInv:
      scale = 1.0/sqrt(fsize);
    break;
    case ConstantAmplitude:
      if(FFTDirection == Forward){
        if(dsize > 0){
          scale = 1.0 / dsize;
        } else {
          brc = false;
        }
      } else {
        scale = dsize / fsize;
      }
    break;
    case ConstantPower:
      scale = 1.0 / sqrt(fsize);
    break;
    default:
    break;
  }
  if(brc){
    if(scale != 1.0){
      for(int i=0; i<FFTSize; i++){
        out[i].re *= scale;
        out[i].im *= scale;
      }
    }
  }
  return brc;
}


// Complex in-place FFT, for non-power-of-two fft sizes.
// This is actually a DFT, not an FFT, so it is VERY slow!
bool __fastcall TCFFT::ipFFT_very_slow(Complex* c1){
  Complex *ctemp = new Complex[FFTSize];
  if(!ctemp) return false; // error!

  Complex *cphasors = new Complex[FFTSize];
  if(!cphasors){
    delete ctemp;
    return false; // error!
  }

  double phs = 2.0 * pi_const / (double)FFTSize;
  for(int i=0; i<FFTSize; i++){
    cphasors[i] = Complex(cos(phs*(double)i), -sin(phs*(double)i));
  }

  for(int i=0; i<FFTSize; i++){
    ctemp[i] = c1[i];
    c1[i] = 0.0;
  }

  for(int i=0; i<FFTSize; i++){
    for(int j=0; j<FFTSize; j++){
      c1[i] = c1[i] + ctemp[j] * cphasors[(i*j)%FFTSize];
    }
  }

  delete ctemp;
  delete cphasors;
  return true;
}

// Complex in-place FFT
bool __fastcall TCFFT::ipFFT_slow(Complex* c1){
   // Determine if FFTSize is a power of two
   if(FFTSize < 1.0e9){ // prevent endless loop
     long powtwo = 2;
     while(powtwo < FFTSize) powtwo *= 2;
     if(powtwo != FFTSize){
       // Can't even use a normal FFT, need to use a DFT!
       return ipFFT_very_slow(c1);
     }
   }

   int n = FFTSize;
   Complex t, w;
   float a,el;
   int mr,i,ii,j,jj,m,nn,l,istep;
   nn=n-1;
   mr=0;

   for(m=1;m<=nn;m++) {
      l=n/2;
      while((mr+l)>nn) {
        l=l/2;
      }
      mr=(mr%l)+l;
      if(mr > m) {
         t = c1[m];
         c1[m] = c1[mr];
         c1[mr] = t;
      }
   }
   l=1;
   while (l<n) {
      istep=l+l;
      el=l;
      for(m=1;m<=l;++m) {
         a=pi_const*(1-m)/el;
         w=Complex(cos(a), sin(a));
         i=m;
         while(i<=n){
            j=i+l;
            jj=j-1;
            ii=i-1;
            t=w*c1[jj];
            c1[jj] = c1[ii] - t;
            c1[ii] += t;
            i=i+istep;
         }
      }
      l=istep;
   }

//   for(i=0; i<n; i++){
//     c1[i] /= n;
//   }
  return true;
}


bool __fastcall TCFFT::CurrentWeightingValid(){
  bool rc = true;
  if(CurrentWeightingSize != FDataSize) rc = false;
  if(CurrentFFTWeighting  != FFTWeighting)     rc = false;
  if(rc && (FFTWeighting == Taylor)){
    if(SidelobeLevel != CurrentSidelobeLevel) rc = false;
    if(TaylorN != CurrentTaylorN) rc = false;
  }
  return rc;
}


bool __fastcall TCFFT::CreateWeighting(){
  if(FDataSize < 2){
    FFTWeighting  = NoWeighting;
    CurrentWeightingSize = FDataSize;
    CurrentFFTWeighting  = FFTWeighting;
  }
  if(Weighting) free(Weighting);
  Weighting = (double*)calloc(FDataSize, sizeof(double));
  if(!Weighting) return false;

  if(FFTWeighting == Hamming){
    for(int i=0; i<FDataSize; i++){
      float a = -pi_const + (double)i / (double)(FDataSize-1)*2*pi_const;
      Weighting[i] = 0.54 + 0.46*cos(a);
    }
  } else if(FFTWeighting == Hanning){
    for(int i=0; i<FDataSize; i++){
      float a = -pi_const + (double)i / (double)(FDataSize-1)*2*pi_const;
      Weighting[i] = 0.5 + 0.5*cos(a);
    }
  } else if(FFTWeighting == Taylor){
    TaylorWeights(Weighting, SidelobeLevel, TaylorN, FDataSize);
  } else if(FFTWeighting == Cheby){
    ChebyWeights(Weighting, SidelobeLevel, FDataSize);
  }

  CurrentWeightingSize = FDataSize;
  CurrentFFTWeighting  = FFTWeighting;

  return true;
}


bool __fastcall TCFFT::TaylorWeights(double *window, double sidelobes, int nbar, int N){
// double *window   = address of array where the Taylor window will be stored.
// double sidelobes = dB level that first set of sidelobes are below the the passband
// int nbar         = The degree to which this window approximates
//                    a Dolph-Chebyschev window.  As nbar increases
//                    this match improves.  The tradeoff is the
//                    number of computataions needed. By the way,
//                    nbar should be greater than A*A+5/4
// int N            = The number of points in the Taylor window.

  double *t = new double[N];
  if(!t) return false;
  double *h = new double[N];
  if(!h) return false;
  double *F = new double[N];
  if(!F) return false;
  for(int i=0; i<N; i++) t[i] = 0.5 + (double)i/(double)(N-1);
  double sll = pow(10.0,sidelobes/20.0);
  double A = log(sll+sqrt(sll*sll-1.0))/pi_const;
  double S = nbar/sqrt(A*A + ((double)nbar-0.5)*((double)nbar-0.5));
  double *k = new double[N];
  if(!k) return false;
  int nbarm = nbar-1;
  for(int i=0; i<N; i++) h[i] = 1.0;
  for(int i=0; i<N; i++) F[i] = 0.0;
  double sign = 1;
  for(int n=0; n<nbarm; n++){
    for(int m=0; m<nbarm; m++){
      double g = (1 - ((n+1)*(n+1))/(S*S*(A*A + ((double)(m+1)-0.5)*((double)(m+1)-0.5))));
      h[n] = g*h[n];
    }
    for(int m=0; m<nbarm; m++) k[m] = 1.0;
    for(int m=0; m<nbarm; m++){
      if(m != n){
        double g = (1-((double)(n+1)/(double)(m+1))*((double)(n+1)/(double)(m+1)));
        k[n] = g*k[n];
      }
    }
    F[n] = sign * h[n]/(2*k[n]);
    sign = -sign;
  }
  for(int i=0; i<N; i++){
    double sumn = 0.0;
    double sumd = 0.0;
    for(int j=0; j<nbarm; j++){
      sumn += F[j] * cos((j+1)*2*pi_const*t[i]);
      sumd += F[j];
    }
    window[i] = (1.0+2.0*sumn)/(1 + 2*sumd);
  }
  delete[] t;
  delete[] h;
  delete[] F;
  delete[] k;

  return true;
}

double acosh(double x){
  if(x > 1) {
    return log(x + sqrt(x*x-1.0));
  } else {
    return 0.0;
  }
}

double Tn(double x, int N){
  if(fabs(x) <= 1) {
    return cos((double)N*acos(x));
  } else {
    return cosh((double)N*acosh(x));
  }
}

bool __fastcall TCFFT::ChebyWeights(double *window, double sidelobes, int N){
// Dolph-Chebyshev window
//
// This routine computes a Chebyshev window.
//   N         = input:  The size of the window
//   sidelobes = input:  The peak sidelobe level, in dB.  This should be a positive value!
//   window    = output: The Chebyshev window weights; a real array of N points.

  double S = pow(10.0,sidelobes/20.0);
  double Zo = cosh(acosh(S)/((double)N-1.0));
  for(int i=0; i<N; i++) window[i] = 0.0;
  int L;
  if(N%2 == 1){
    L = (N-1)/2;
  } else {
    L = N/2 - 1;
  }
  for(int i=0; i<N; i++){
    double suml = 0.0;
    for(int n=0; n<L; n++){
      double arg1 = Zo * cos(pi_const*((double)n+1.0)/(double)N);
      double arg2 = (2*pi_const*(double)(n+1)/(double)N) * ((i+1)-((double)N+1.0)/2.0);
      suml += Tn(arg1, N-1)*cos(arg2);
    }
    window[i] = (2.0/(double)N) * (S+2*suml);
  }
  double maxwgt = window[0];
  for(int i=1; i<N; i++){
    if(window[i] > maxwgt) maxwgt = window[i];
  }
  for(int i=0; i<N; i++){
    window[i] /= maxwgt;
  }
  return true;
}

